(*
 *  beware IRC services, bchannel.pas
 *  Copyright (C) 2002 Bas Steendijk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *)

unit bchannel;

interface

uses blinklist,buser,bstuff,bconsts,bcmds;

const
  chanmode_inviteonly= $1;
  chanmode_moderated=  $2;
  chanmode_noexternal= $4;
  chanmode_private=    $8;
  chanmode_secret=    $10;
  chanmode_topic=     $20;
  chanmode_reggedonly=$40;
  chanmode_nocolors=  $80;
  chanmode_noctcp=   $100;
  chanmode_delayedjoin=   $200;

  userchanflag_voice=  $1;
  userchanflag_op=     $2;

  chanflag_destroying=$1000;
  chanflag_modeless=  $2000;
  chanflag_local=     $4000;

  chanflag_nameset=   $8000;

  schanmodesupported='biklmnopstvr';
  isupportchantypes='+#&';
  sisupportchanmodes='b,k,l,imnpstr';
  isupportprefix='(ov)@+';

type
  {one ban}
  tban=class(tlinklist)
    mask:string;
    umask:string;
    date:integer;
    sender:string;
  end;

  tchannel=class(tthing)

    modeflag:integer;
    flags:integer;
    limit:integer;
    key:string;
    user:tlinklist2;
    banlist:tban;
    bancount:integer;
    topic:string;
    topicby:string;
    topictime:integer;
    localuser:tplinklist;
    usercount:integer;
    ts:integer;

    {# of non-deaf users behind this link, used for chat routing}

    destructor destroy; override;
  end;

  {channel-user relationship. prev,next are user's channel list. prev2,next2 are channel's user list}
  tuserchan=class(tlinklist2)
    ch:tchannel;
    us:tuser;
    flags:integer; {ops, voice}
    lu:tplinklist; {local-user struct, if any}
  end;

var
  globalchanlist:tlinklist;

var
  chanmodetable_reggedonly,
  chanmodetable_nocolors,chanmodetable_noctcp,
  chanmodetable_delayedjoin,
  chanmodetable_inviteonly,chanmodetable_moderated,
  chanmodetable_noexternal,chanmodetable_private,
  chanmodetable_secret,chanmodetable_topic:integer;

type
  tchanmodetable=record
    c:char;               {mode char}
    flag:integer;         {mode flag}
(*    disabled:boolean;     {disabled by options (cant be seen/used by clients)}
    auto:boolean;         {no special precautions needed when changing the mode, by a client}*)
    num:^integer;
  end;

const
  maxchanmodetable=9;

  chanmodetable:array[0..maxchanmodetable] of tchanmodetable=(
  (c:'r';flag:chanmode_reggedonly;num:@chanmodetable_reggedonly),
  (c:'c';flag:chanmode_nocolors;num:@chanmodetable_nocolors),
  (c:'C';flag:chanmode_noctcp;num:@chanmodetable_noctcp),
  (c:'D';flag:chanmode_delayedjoin;num:@chanmodetable_delayedjoin),
  (c:'i';flag:chanmode_inviteonly;num:@chanmodetable_inviteonly),
  (c:'m';flag:chanmode_moderated;num:@chanmodetable_moderated),
  (c:'n';flag:chanmode_noexternal;num:@chanmodetable_noexternal),
  (c:'p';flag:chanmode_private;num:@chanmodetable_private),
  (c:'s';flag:chanmode_secret;num:@chanmodetable_secret),
  (c:'t';flag:chanmode_topic;num:@chanmodetable_topic)
  );


function ischanprefix(c:char):boolean;
function validchanname(const s:string):boolean;

{return ch for given name, nil if not exists, s does not need to be uppercased}
function findchan(const s:string):tchannel;

{
change the name of a channel
channel's names don't change, but a routine like this
better allows things like a search tree
}
procedure setchanname(ch:tchannel;const s:string);

{create user/channel relation (but don't send anything)  returns the chanuser}
function addusertochannel(us:tuser;ch:tchannel):tuserchan;

{remove user/channel relation,
if cu is not nil, it is used instead of a slower seacrh
if channel became empty, it is destroyed

the user must be on the channel (no checking done here)
}
procedure deluserfromchannel(us:tuser;ch:tchannel;uc:tuserchan);

{make a new channel, add it to list}
function createchannel:tchannel;

{returns true if us is member of ch. allows parameters to be nil}
function isonchannel(us:tuser;ch:tchannel):boolean;

function getuserchan(us:tuser;ch:tchannel):tuserchan;

{if user can see channel (in whois etc)}
function canseechannel(us:tuser;ch:tchannel):boolean;
function issecret(ch:tchannel):boolean;
function isprivate(ch:tchannel):boolean;
function ismoderated(ch:tchannel):boolean;
function isnoexternal(ch:tchannel):boolean;

function hasops(us:tuser;ch:tchannel;uc:tuserchan):boolean;
function hasvoice(us:tuser;ch:tchannel;uc:tuserchan):boolean;

function chanmodestr(ch:tchannel;params:boolean):string;

{adds a ban to the channel. if the ban overlaps existing bans,
theyll be removed and a modechange sent to local users}
procedure addban(sender:tuser;ch:tchannel;const s:string);

{returns pointer to ban or nil if the ban is not found}
function findban(ch:tchannel;mask:string):tban;

procedure setchanmode(ch:tchannel;cptr,sptr:tuser;parc:integer;parv:pparams);


implementation

uses sysutils,bsearchtree,btime,bconfig;

var
  chantree:thashtable;


function ischanprefix(c:char):boolean;
begin
  result := (c = '#') or (c = '&') or (c = '+');
end;

function validchanname(const s:string):boolean;
var
  a:integer;
begin
  result := false;
  if s = '' then exit;
  if length(s) > maxchanname then exit;
  if not ((s[1] = '#') or (s[1] = '&') or (s[1] = '+')) then exit;
  if length(s) > 1 then for a := 2 to length(s) do begin
    if (s[a] = #7) or (s[a] = ',') or (s[a] = #0) then exit
  end;
  result := true
end;

destructor tchannel.destroy;
var
  p:tban;
begin
  if flag_isset(self.flags,chanflag_destroying) then exit;
  setflag(self.flags,chanflag_destroying);

  while self.user <> nil do deluserfromchannel(tuserchan(self.user).us,self,nil);

  while self.banlist <> nil do begin
    p := self.banlist;
    linklistdel(tlinklist(self.banlist),tlinklist(self.banlist));
    p.destroy;
  end;

  setchanname(self,'');  {de-init channel name}
  linklistdel(globalchanlist,tlinklist(self));
  dec(count.channels);
  inherited destroy;
end;


function findchan(const s:string):tchannel;
begin
  result := findtree(@chantree,ircupper(s));
end;

function addusertochannel(us:tuser;ch:tchannel):tuserchan;
var
  uc:tuserchan;
  lu:tplinklist;
begin
  uc := tuserchan.create;

  uc.ch := ch;
  uc.us := us;

  linklistadd(us.channel,tlinklist(uc));
  linklist2add(ch.user,tlinklist2(uc));

  inc(us.chancount);
  inc(ch.usercount);

  if myconnect(us) then begin
    lu := tplinklist.create;
    lu.p := us;
    uc.lu := lu;
    linklistadd(tlinklist(ch.localuser),tlinklist(lu));
  end;
  result := uc;
end;

procedure deluserfromchannel(us:tuser;ch:tchannel;uc:tuserchan);
begin
  if uc = nil then begin
    {search userchan}
    uc := tuserchan(us.channel);
    while uc <> nil do begin
      if uc.ch = ch then begin
        break;
      end;
      uc := tuserchan(uc.next);
    end;
  end;

  if uc.lu <> nil then begin
    linklistdel(tlinklist(ch.localuser),tlinklist(uc.lu));
    uc.lu.destroy
  end;

  linklistdel(us.channel,uc);
  linklist2del(ch.user,uc);
  uc.destroy;
  dec(us.chancount);
  dec(ch.usercount);

  if ch.usercount <= 0 then ch.destroy;
end;


function createchannel;
begin
  result := tchannel.create;
  linklistadd(globalchanlist,tlinklist(result));
  inc(count.channels);
end;


procedure setchanname(ch:tchannel;const s:string);
begin
  if flag_isset(ch.flags,chanflag_nameset) then deltree(@chantree,ircupper(ch.name));
  ch.name := s;
  if ch.name <> '' then begin
    addtree(@chantree,ircupper(ch.name),ch);
    setflag(ch.flags,chanflag_nameset)
  end else
  clearflag(ch.flags,chanflag_nameset);
end;

function isonchannel(us:tuser;ch:tchannel):boolean;
begin
  result := getuserchan(us,ch) <> nil;
end;

function getuserchan(us:tuser;ch:tchannel):tuserchan;
var
  p2:tlinklist2;
begin
  result := nil;
  if (ch = nil) or (us = nil) then exit;
  if us.chancount <= ch.usercount then begin
    {faster to check user for all channels}
    p2 := tlinklist2(us.channel);
    while p2 <> nil do begin
      if tuserchan(p2).ch = ch then begin
        result := tuserchan(p2);
        exit;
      end;
      p2 := tlinklist2(p2.next);
    end;
  end else begin
    {faster to check channel for all users}
    p2 := ch.user;
    while p2 <> nil do begin
      if tuserchan(p2).us = us then begin
        result := tuserchan(p2);
        exit;
      end;
      p2 := p2.next2;
    end;
  end;
end;

function issecret(ch:tchannel):boolean;
begin
  result := flag_isset(ch.modeflag,chanmode_secret);
end;

function isprivate(ch:tchannel):boolean;
begin
  result := flag_isset(ch.modeflag,chanmode_private);
end;

function ismoderated(ch:tchannel):boolean;
begin
  result := flag_isset(ch.modeflag,chanmode_moderated);
end;

function isnoexternal(ch:tchannel):boolean;
begin
  result := flag_isset(ch.modeflag,chanmode_noexternal);
end;

function hasops(us:tuser;ch:tchannel;uc:tuserchan):boolean;
begin
  result := false;
  if uc = nil then begin
    uc := tuserchan(us.channel);
    while uc <> nil do begin
      if uc.ch = ch then begin
        break;
      end;
      uc := tuserchan(uc.next);
    end;
  end;
  if uc = nil then exit;
  result := flag_isset(uc.flags,userchanflag_op);
end;

function hasvoice(us:tuser;ch:tchannel;uc:tuserchan):boolean;
begin
  result := false;
  if uc = nil then begin
    uc := tuserchan(us.channel);
    while uc <> nil do begin
      if uc.ch = ch then begin
        break;
      end;
      uc := tuserchan(uc.next);
    end;
  end;
  if uc = nil then exit;
  result := flag_isset(uc.flags,userchanflag_voice);
  {}
end;


function canseechannel(us:tuser;ch:tchannel):boolean;
begin
  if not (flag_isset(ch.modeflag,chanmode_secret) or
  flag_isset(ch.modeflag,chanmode_private)) then begin
    result := true;
    exit;
  end;
  result := isonchannel(us,ch);
end;

function chanmodestr(ch:tchannel;params:boolean):string;
var
  s2:string;
  a:integer;
begin
  if (ch.modeflag = 0) and (ch.key = '') and (ch.limit = 0) then begin
    result := '';
    exit;
  end;
  result := ' +';
  s2 := '';
  for a := 0 to maxchanmodetable do begin
    if ch.modeflag and chanmodetable[a].flag <> 0 then result := result + chanmodetable[a].c;
  end;
  if ch.key <> '' then begin
    result := result + 'k';
    s2 := s2 + ' '+ch.key;
  end;
  if ch.limit > 0 then begin
    result := result + 'l';
    s2 := s2 + ' '+inttostr(ch.limit);
  end;
  if params then result := result + s2;
end;

procedure addban(sender:tuser;ch:tchannel;const s:string);
var
  s2:string;
  b,b2:tban;
begin
  s2 := ircupper(s);

  {remove overlapped bans}

  b := ch.banlist;
  while b <> nil do begin
    b2 := tban(b.next);
    if maskmatch(s2,b.umask) then begin
      linklistdel(tlinklist(ch.banlist),tlinklist(b));
      b.destroy;
      dec(ch.bancount);
    end;
    b := b2;
  end;

  b := tban.create;
  b.mask := s;
  b.umask := s2;
  b.sender := sender.name;
  b.date := irctime;
  linklistadd(tlinklist(ch.banlist),tlinklist(b));
  inc(ch.bancount);
end;

function findban(ch:tchannel;mask:string):tban;
var
  b:tban;
begin
  result := nil;
  mask := ircupper(mask);
  b := ch.banlist;
  while b <> nil do begin
    if b.umask = mask then begin
      result := b;
      exit;
    end;
    b := tban(b.next);
  end;
end;

function charhasparam(c:char;setclear:integer):boolean;
begin
  case c of
    'b','k','o','v':result := true;
    'l':result := setclear = 2;
    else result := false
  end;
end;


procedure setchanmode(ch:tchannel;cptr,sptr:tuser;parc:integer;parv:pparams);
var
  cc,pc:integer;
  a,setclear:integer;
  s,s2:string;
  ban:tban;

{+n}
procedure setmodechar(k:char;flag:integer;setclear:integer);
begin
  if setclear = 2 then begin
    ch.modeflag := ch.modeflag or flag;
  end else begin
    ch.modeflag := ch.modeflag and not flag;
  end;
end;

{+o user}
procedure setmodechannick(k:char;flag:integer;setclear:integer;nick:string);
var
  us:tuser;
  uc:tuserchan;
begin
  us := findnumeric(nick);
  if us = nil then exit;
  uc := getuserchan(us,ch);
  if uc = nil then exit;
  if setclear = 2 then begin
    uc.flags := uc.flags or flag
  end else begin
    uc.flags := uc.flags and not flag
  end;
end;

begin
  s := parv[2];
  setclear := 2;
  pc := 3;
  parv[3+maxmodes] := ''; {trick to have max. 6 modes with parameters}
  for cc := 1 to length(s) do begin
    case s[cc] of
      '-':setclear := 1;
      '+':setclear := 2;
      else begin
        case s[cc] of
          'b':begin
            if parv[pc] <> '' then begin
              s2 := copy(cookmask(parv[pc]),1,maxbanlength);
              ban := findban(ch,s2);
              if setclear = 2 then begin
                addban(sptr,ch,s2);
              end else begin
                if ban <> nil then begin
                  linklistdel(tlinklist(ch.banlist),tlinklist(ban));
                  ban.destroy;
                  dec(ch.bancount);
                end;
              end;
            end;
          end;
          'o':setmodechannick('o',userchanflag_op,setclear,parv[pc]);
          'v':setmodechannick('v',userchanflag_voice,setclear,parv[pc]);
          'k':if pc < parc then begin
            s2 := '';
            for a := 1 to length(parv[pc]) do if parv[pc,a] > #32 then s2 := s2 + parv[pc,a];
            s2 := copy(s2,1,maxkeylength);
            if s2 <> '' then begin
              if setclear = 2 then begin
                ch.key := s2;
              end else begin
                ch.key := '';
              end;
            end;
          end;
          'l':begin
            if setclear = 2 then begin
              a := strtointdef(parv[pc],0);
              ch.limit := a;
            end else begin
              ch.limit := 0;
            end;
          end;
        else
          for a := 0 to maxchanmodetable do if chanmodetable[a].c = s[cc] then
          setmodechar(chanmodetable[a].c,chanmodetable[a].flag,setclear);

        end;
      end;
    end;
    if charhasparam(s[cc],setclear) then begin
      inc(pc);
      if pc > 3+maxmodes then pc := 3+maxmodes;
    end;
  end;
  {parse TS}
  if parc > pc then begin
    a := strtointdef(parv[pc],0);
    if a > 0 then if a < ch.ts then ch.ts := a;
  end;
end;

procedure init;
var
  a:integer;
begin
  globalchanlist := nil;
  for a := 0 to maxchanmodetable do if chanmodetable[a].num <> nil then
  chanmodetable[a].num^ := a;
end;

initialization init;

end.

