(*
 *  beware IRC services, bworldaccount.pas
 *  Copyright (C) 2002 Bas Steendijk
 *
 *  modifications by peter green
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *)


unit bworldaccount;

interface

uses blinklist,bstuff,buser;

const
  idnicklen=16;

type
  tbworldaccount=class(tlinklist)
    name:string;
    password:string;
    level:integer;
    mask:tstringlinklist;
    (*user:tuser;  {who is logged in to this account, nil=noone}*)
    user:tplinklist;
    marked:boolean; {internal}
    operonly:boolean;

    function addmask(s:string):boolean;
    function delmask(s:string):boolean;
    function adduser(us:tuser):boolean;
    function removeuser(us:tuser):boolean;
    procedure clearmasks;
    constructor create;
    destructor destroy; override;
  end;

function findbworldaccount(const s:string):tbworldaccount;

{add account. only to be used from command handler or similar}
function addbworldaccount(const name:string;level:integer):tbworldaccount;

function validaccountname(const s:string):boolean;

function convertaccountname(const s:string):string;

procedure clearbworldaccount;
procedure savebworldaccount(const filename:string);
procedure loadbworldaccount(const filename:string);

var
  bworldaccountlist:tbworldaccount;

implementation

uses bworldbot,bconfig,passcryp,bparse,bcmds,sysutils,unitwritetxt;

function validaccountname(const s:string):boolean;
var
  a:integer;
begin
  result := false;
  if (s = '') or (length(s) > idnicklen) then exit;
  for a := 1 to length(s) do begin
    if not ((s[a] in ['A'..'Z']) or (s[a] in ['0'..'9'])) then exit;
  end;
  result := true;
end;

{
centralized code to convert (or not) to uppercase
}
function convertaccountname(const s:string):string;
begin
  result := ircupper(s);
end;
{
IDnick's are case tolerant here
i may later make hashtable code for this,
but there arent thousands of accounts anyway, so this will do for now
}
function findbworldaccount(const s:string):tbworldaccount;
var
  p:tbworldaccount;
begin
  p := bworldaccountlist;
  result := nil;
  while p <> nil do begin
    if strcompup(p.name,s) then begin
      result := p;
      exit
    end;
    p := tbworldaccount(p.next);
  end;
end;


function addbworldaccount(const name:string;level:integer):tbworldaccount;
var
  p:tbworldaccount;
begin
  result := nil;
  if findbworldaccount(name) <> nil then exit; {aready exists}
  p := tbworldaccount.create;
  linklistadd(tlinklist(bworldaccountlist),tlinklist(p));
  p.name := name;
  p.level := level;
  result := p;
end;

function TBWorldAccount.adduser(us:tuser):boolean;
var
  p:tplinklist;
begin
  result := false;
  {search user already in list}
  p := self.user;
  while p <> nil do begin
    if p.p = us then exit;
    p := tplinklist(p.next);
  end;
  result := true;
  p := tplinklist.create;
  p.p := us;
  linklistadd(tlinklist(self.user),tlinklist(p));
end;

function TBWorldAccount.removeuser(us:tuser):boolean;
var
  p:tplinklist;
begin
  result := true;
  {search user in list}
  p := tplinklist(self.user);
  while p <> nil do begin
    if p.p = us then begin
      linklistdel(tlinklist(self.user),tlinklist(p));
      p.destroy;
      exit;
    end;
    p := tplinklist(p.next);
  end;
  result := false;

end;

function tbworldaccount.addmask(s:string):boolean;
var
  s2:string;
  sl:tstringlinklist;

begin
  result := false;

  s := cookmask(s);
  {process mask (like ban), must also check here if mask is valid}

  s2 := ircupper(s);
  sl := self.mask;
  while sl <> nil do begin
    if ircupper(sl.s) = s2 then exit; {mask already exists}
    sl := tstringlinklist(sl.next);
  end;

  sl := tstringlinklist.create;
  sl.s := s;
  linklistadd(tlinklist(self.mask),tlinklist(sl));
  result := true;
end;

function tbworldaccount.delmask(s:string):boolean;
var
  s2:string;
  sl,sl2:tstringlinklist;

begin
  result := false;

  s := cookmask(s);
  {process mask (like ban), must also check here if mask is valid}

  s2 := ircupper(s);
  sl2 := nil;
  sl := self.mask;
  while sl <> nil do begin
    if ircupper(sl.s) = s2 then begin
      sl2 := sl;
      break;
    end;
    sl := tstringlinklist(sl.next);
  end;

  if sl2 = nil then exit; {mask doesn't exist}
  linklistdel(tlinklist(self.mask),tlinklist(sl2));
  sl2.destroy;
  result := true;
end;

constructor tbworldaccount.create;
begin
  inherited create;
  {}
end;

destructor tbworldaccount.destroy;
begin
  while self.user <> nil do bworldbot.deauth(tuser(self.user.p));

  self.clearmasks;
  inherited destroy;
end;

procedure tbworldaccount.clearmasks;
var
  sl,sl2:tstringlinklist;
begin
  {destroy all masks}
  sl := self.mask;
  while sl <> nil do begin
    sl2 := tstringlinklist(sl.next);
    linklistdel(tlinklist(self.mask),tlinklist(sl));
    sl.destroy;
    sl := sl2;
  end;
end;

{free bworld oper account list (to be done on shutdown)}
procedure clearbworldaccount;
var
  p,p2:tbworldaccount;
begin
  p := bworldaccountlist;
  while p <> nil do begin
    p2 := tbworldaccount(p.next);
    linklistdel(tlinklist(bworldaccountlist),tlinklist(p));
    p.destroy;
    p := p2;
  end;
end;

var
  currentaccount:tbworldaccount;

procedure loadbworldaccountproc(sv,sn:string);
begin
  if sv = 'NAME' then begin
    currentaccount := nil;
    if not validaccountname(convertaccountname(sn)) then exit;

    currentaccount := findbworldaccount(sn);
    if currentaccount = nil then begin
      currentaccount := tbworldaccount.create;
      linklistadd(tlinklist(bworldaccountlist),tlinklist(currentaccount));
    end else begin
      currentaccount.clearmasks;
    end;
    currentaccount.name := sn; {if the account already exists, this allows changing the case of the name}
    currentaccount.marked := true;
  end else
  if sv = 'LEVEL' then begin
    if currentaccount <> nil then begin
      currentaccount.level := strtointdef(sn,0);
    end;
  end else if sv = 'MASK' then begin
    if currentaccount <> nil then begin
      sn := cookmask(sn);
      if validmask(sn) then currentaccount.addmask(sn);
    end;
  end else if sv = 'PASSWORD' then begin
    if currentaccount <> nil then begin
      if iscryptedlength(length(sn)) or (sn = '') then
      currentaccount.password := sn
      else
      currentaccount.password := passnewcrypt(sn);
    end;
  end else if sv = 'OPERONLY' then begin
    if currentaccount <> nil then begin
      currentaccount.operonly := sn = '1';
    end;
  end

end;


procedure loadbworldaccount(const filename:string);
var
  p,p2:tbworldaccount;

begin
  //writeln('loadbworldaccount called');
  {don't clear list}

  {un-mark all accounts}
  p := bworldaccountlist;
  while p <> nil do begin
    p.marked := false;
    p := tbworldaccount(p.next);
  end;

  currentaccount := nil;
  readcfg(filename,@loadbworldaccountproc);
  currentaccount := nil;

  p := bworldaccountlist;
  while p <> nil do begin
    p2 := tbworldaccount(p.next);
    if not p.marked then begin
      linklistdel(tlinklist(bworldaccountlist),tlinklist(p));
      p.destroy;
    end;
    p := p2;
  end;
end;



procedure savebworldaccount(const filename:string);
var
  t:textfile;
  sl:tstringlinklist;
  p:tbworldaccount;
begin
  if not openforwrite(t,filename) then exit;
  p := tbworldaccount(bworldaccountlist);
  while p <> nil do begin
    {}
    writeln(t,'[account]');
    writeln(t,'name='+p.name);
    writeln(t,'level='+inttostr(p.level));
    writeln(t,'password='+p.password);
    writeln(t,'operonly='+inttostr(ord(p.operonly)));
    sl := p.mask;
    while sl <> nil do begin
      writeln(t,'mask='+sl.s);
      sl := tstringlinklist(sl.next);
    end;
    p := tbworldaccount(p.next);
    writeln(t);
  end;
  closefile(t);
end;

end.
