(*
 *  beware IRC services, unitbanmask.pas
 *  Copyright (C) 2002 Bas Steendijk
 *
 *  some modifications by peter green
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *)
unit unitbanmask;

interface

uses bstuff,bconsts,bpremaskmatch;

{
unit to speed up banlist/G-list checking by doing binary match if possible

binary IP adresses and masks are stored as cpu byte order


}

type
  tbanmask=record
    notip:boolean;
    bin:boolean;
    ip:integer;
    bits:integer;
    nouserid:boolean;
    user:string;
    uuser:string;
    host:string;
    uhost:string;
  end;
  pbanmask=^tbanmask;

procedure banmaskmake(p:pbanmask;s:string);
function banmaskmatch(p1,p2:pbanmask):boolean;
function banpremaskmatch(premask:ppremask;p1,p2:pbanmask;s1,s2:string):boolean;
function getmaskstr(p:tbanmask):string;
procedure banmaskmake_oneuser(p:pbanmask;userid,host:string;ip:integer);

implementation

procedure banmaskmake(p:pbanmask;s:string);
label nobin;
var
  a,b,d:integer;
  goed:boolean;
  s2,s3:string;
begin
  {separate userid and host}
  a := pos('@',s);
  if a = 0 then begin
    p.nouserid := true;
  end else begin
    p.user := copy(s,1,a-1);
    s := copy(s,a+1,hostlen);
    p.nouserid := p.user = '*';
  end;

  p.uuser := ircupper(p.user);
  p.host := s;
  p.uhost := ircupper(p.host);

  (*
  if pos('/',s) <> 0 then begin
    {block/bits notation, if fails, will be string mask}
  end;
  *)

  {normal ip mask or single ip. if fails, will be string mask}
  goed := true;
  b := 24;

  s2 := '';
  p.ip := 0;
  s3 := s + '.';
  p.bits := 0;
  for a := 1 to length(s3) do begin
    case s3[a] of
      '.':begin
        if b <= 0 then if not ((b = 0) and (a = length(s3))) then begin
          {failed, too many dots}
          goed := false;
          break;
        end;
        if s2 = '*' then begin
          {}
        end else begin
          d := strtointdef(s2,-1);
          if (d < 0) or (d > 255) then begin
            {failed, invalid number}
            goed := false;
            break;
          end;
          p.bits := p.bits or ($ff shl b);
          p.ip := p.ip or (d shl b);
        end;
        b := b-8;
        s2 := '';
      end;
      else begin
        s2 := s2 + s[a];
      end;
    end;
  end;

  {if there are less than 3 dots in the mask, only the last char may be *}
  if goed then if b <> -8 then if pos('*',s3) <> (length(s3) -1) then goed := false;

  if goed then goed := (s[length(s)] = '*') or (b = -8);
  if goed then begin
    p.bin := true;
    p.ip := p.ip and p.bits;
    exit;
  end;

nobin:
{string mask, no binary}
  p.notip := not isipmask(s);

  p.bin := false;
  p.ip := 0;
  p.bits := 0;
end;

procedure banmaskmake_oneuser(p:pbanmask;userid,host:string;ip:integer);
begin
  p.bin := ip <> 0;
  p.ip := ip;
  p.bits := $ffffffff;
  p.host := host;
  p.uhost := ircupper(host);
  p.nouserid := false;
  p.user := userid;
  p.uuser := ircupper(userid);
end;


{
check if p1 overlaps p2
}

function banmaskmatch(p1,p2:pbanmask):boolean;
begin
  if (p1.bin) and (p2.bin) then begin
    result := p1.ip = (p2.ip and p1.bits);
  end else begin
    result := maskmatch(p1.uhost,p2.uhost);
  end;
  if not result then exit;
  if p1.nouserid then exit;
  result := maskmatch(p1.uuser,p2.uuser);
end;

function banpremaskmatch(premask:ppremask;p1,p2:pbanmask;s1,s2:string):boolean;
begin
  if p1.bin then begin
    if p2.bin then begin
      result := banmaskmatch(p1,p2);
    end else begin
      result := premaskmatch(premask,s2);
      if result then result := maskmatch(s1,s2);
    end;
  end else begin
    if p2.bin and p1.notip then begin
      result := false;
    end else begin
      result := premaskmatch(premask,s2);
      if result then result := maskmatch(s1,s2);
    end;
  end;
end;

function getmaskstr(p:tbanmask):string;
begin
  if p.nouserid then result := '*' else result := p.user;
  result := result + '@'+p.host;
end;


end.


